﻿using FObjBase;
using GeneralGommunication;
using Newtonsoft.Json;
using NLog;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.IO;
using System.Linq;

namespace CapSensorBase
{
    public class CapSensor2 : ICapSensor2Client
    {
        public Logger logger = NLog.LogManager.GetCurrentClassLogger();

        public event PropertyChangedEventHandler PropertyChanged;

        #region MARKNO
        const int MARKNO_SET_SAVE = 4;
        #endregion

        #region  ICapSensor2

        public event HzMeasuredOutEventHandler HzMeasuredOut;
        public event IStatusChangedEventHandler IStatusChanged;
        public event Action<object> SystickReseted;
        /// <summary>
        /// 获取 systick 转换的时间
        /// </summary>
        public DateTime Now { get; private set; }

        /// <summary>
        /// 通道值，只有16位,更新频率为10ms
        /// </summary>
        public UInt16 Data { get; private set; }

        /// <summary>
        /// 1s 频率
        /// </summary>
        public double KHz { get; private set; }
        /// <summary>
        /// 输入口状态，只有1位, 更新频率为1ms
        /// </summary>
        public UInt16 IStatus { get; private set; }

        #endregion

        #region ICapSensor2Client

        /// <summary>
        /// 运行中
        /// </summary>
        public bool IsRunning { get; private set; }

        /// <summary>
        /// 连接成功
        /// </summary>
        public bool IsConnected { get; protected set; }

        /// <summary>
        /// 重连次数
        /// </summary>
        public int ConnectCnt { get; protected set; }


        /// <summary>
        /// 已经获取全部当前数据
        /// </summary>
        public bool IsReady { get; protected set; }

        /// <summary>
        ///  通过判断systick，判断AD盒是否发生重启;
        ///  BeResetTime为上一次被复位时间， 此为 电脑的时间
        /// </summary>
        public DateTime BeResetTime { get; private set; }

        /// <summary>
        /// AD盒地址
        /// </summary>
        public string Addr { get; set; } = "192.168.251.12:20020";
        #endregion

        public CapSensorCore2 Core => core;

        IGeneralComm comm;
        CapSensorCore2 core;
        IsReadyContext isReadyContext = new IsReadyContext();
        SysTickContext sysTickContext = new SysTickContext();

        /// <summary>
        /// 检查推送周期 10ms 有一个。 5秒内都没收到，应该断开了， 重连
        /// </summary>
        Stopwatch stopwatch_pushIntervalCheck = new Stopwatch();
        class AsyncEvent
        {
            public Action<EventArgs> action;
            public EventArgs e;
        }
        List<AsyncEvent> asyncEvents = new List<AsyncEvent>();

        static CapSensor2()
        {
        }
        public CapSensor2()
        {
            constructor();
        }
        void constructor()
        {
            //Now = DateTime.Now;
            //sysTickContext.BeReseted += () => BeResetTime = DateTime.Now;

            core = new CapSensorCore2();
            core.logger = logger;
            core.Csm.StartMeasure();

            //注册capsensor的发送动作，把数据通过 comm模块发送出去
            core.SendMsgEvent += (s, msg) =>
            {
                comm?.Write(msg);
            };

            core.HzMeasuredOut += Core_HzMeasuredOut;
            core.IStatusChanged += Core_IStatusChanged;

            //避免推送太快，全部推送事件统一周期处理
            PollModule.Current.Poll_Config(onPoll_AsyncEvent);
            PollModule.Current.Poll_Config(onPoll_CheckPush);

            Misc.BindingOperations.SetBinding(core, nameof(core.IsConnected), this, nameof(IsConnected));
            Misc.BindingOperations.SetBinding(isReadyContext, nameof(isReadyContext.IsReady), this, nameof(IsReady));

            this.PropertyChanged += CapSensor_PropertyChanged;
        }

        private void onPoll_CheckPush()
        {
            if (!IsConnected)
            {
                if (stopwatch_pushIntervalCheck.IsRunning)
                    stopwatch_pushIntervalCheck.Stop();
                return;
            }

            if (!stopwatch_pushIntervalCheck.IsRunning)
            {
                stopwatch_pushIntervalCheck.Restart();
            }
            else
            {
                if (stopwatch_pushIntervalCheck.ElapsedMilliseconds > 5000)
                {
                    //5000ms 都没收到推送，连接断开
                    logger.Error("5000ms 都没收到推送 主动重连");
                    ReConnect();
                }
            }
        }

        private void Core_IStatusChanged(object sender, CoreEvent_IStatusChangedEventArgs e)
        {
            SetNow(e.systick);
            IStatus = e.istatus;

            AddToPushEventArgsList(new AsyncEvent()
            {
                action = (_e) =>
                {
                    IStatusChanged?.Invoke(this, _e as IStatusChangedEventArgs);
                },
                e = new IStatusChangedEventArgs()
                {
                    time = Now,
                    istatus = e.istatus,
                    ichanged = e.ichanged
                }
            });
        }
        void AddToPushEventArgsList(AsyncEvent e)
        {

            //这个线程非主线程， 数据接收完，应该快速返回。
            //事件触发，都放在主线程操作。
            //要有线程锁！！！！！
            lock (asyncEvents)
            {
                asyncEvents.Add(e);
            }
        }
        /// <summary>
        /// 推送的最大间隔ms
        /// </summary>
        public int PushIntervalMsMax { get; set; }

        /// <summary>
        /// 推送的平均间隔ms
        /// </summary>
        public int PushIntervalMsAvg { get; set; }

        /// <summary>
        /// 推送间隔ms 异常次数
        /// </summary>
        public int PushIntervalMsErrCnt { get; set; }

        /// <summary>
        /// systick 不连续次数
        /// </summary>
        public int SystickErrCnt { get; set; }

        List<int> pushIntervalMsList = new List<int>();
        byte? last_systick = null;
        private void Core_HzMeasuredOut(object sender, CoreEvent_HzMeasuredOutEventArgs e)
        {
            bool isSystickReseted = false;
            if (last_systick != null)
            {
                byte delta = (byte)(e.systick - (byte)last_systick);
                if (delta > 50)
                {
                    //两次数据包，systick间隔正常是10ms  大于 15ms肯定是掉包

                    SystickErrCnt++;
                    //复位时间
                    sysTickContext.Reset();
                    isSystickReseted = true;
                }
            }
            last_systick = e.systick;

            int ms = (int)stopwatch_pushIntervalCheck.ElapsedMilliseconds;
            pushIntervalMsList.Add(ms);
            if (pushIntervalMsList.Count() > 500)
            {
                //正常情况10ms一次， 500次就是5s
                PushIntervalMsMax = pushIntervalMsList.Max();
                PushIntervalMsAvg = (int)Math.Round(pushIntervalMsList.Average());
                pushIntervalMsList.Clear();
            }

            //喂狗
            stopwatch_pushIntervalCheck.Restart();

            //isReady = true
            isReadyContext.isReadyGetIn = true;

            if (ms > 500)
            {
                //电容传感器 10ms推送一次，但 FLY 422模块会收满150byte 才推送， 最后到PC 的间隔大约为 200ms
                //当两次接收大于500ms， 肯定连接断开了

                PushIntervalMsErrCnt++;

                //肯定异常需要重新同步

                //复位时间
                sysTickContext.Reset();
                isSystickReseted = true;
            }
            SetNow(e.systick);
            Data = e.data;

            updateKHz();

            AddToPushEventArgsList(new AsyncEvent()
            {
                action = (_e) =>
                {
                    var e2 = _e as HzMeasuredOutEventArgs;
                    HzMeasuredOut?.Invoke(this, e2);

                    if (e2.isSystickReseted)
                    {
                        //通知同步器执行同步
                        SystickReseted?.Invoke(this);
                    }
                },
                e = new HzMeasuredOutEventArgs()
                {
                    time = Now,
                    data = Data * 100, //正常10ms一次 所以需要 x100
                    isSystickReseted = isSystickReseted
                }
            });
        }


        /// <summary>
        /// 低频处理事件
        /// </summary>
        void onPoll_AsyncEvent()
        {
            AsyncEvent[] asyncEvents;
            lock (this.asyncEvents)
            {
                asyncEvents = this.asyncEvents.ToArray();
                this.asyncEvents.Clear();
            }

            foreach (var asyncEvent in asyncEvents)
            {
                asyncEvent.action(asyncEvent.e);
            }

        }


        void SetNow(byte systick)
        {
            Now = sysTickContext.ToDateTime(systick);
        }


        DateTime dataListBeginTime;
        List<int> dataList = new List<int>();

        void updateKHz()
        {
            if (dataList.Count() == 0)
            {
                dataListBeginTime = Now;
            }
            dataList.Add(Data);
            int ms = (int)(Now - dataListBeginTime).TotalMilliseconds;
            if (ms >= 490)
            {
                double sum = dataList.Sum(d => d);
                KHz = sum * 100 / dataList.Count() / 1000.0;
                dataList.Clear();
            }
        }

        private void CapSensor_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName == nameof(IsConnected))
            {
                if (IsConnected)
                {
                    ConnectCnt++;
                    BeResetTime = DateTime.Now;
                    AfterConnected();
                }
                else
                {
                    isReadyContext.Reset();
                }
            }
        }
        /// <summary>
        /// 连接后初始化
        /// </summary>
        void AfterConnected()
        {
            core.logger.Debug($"CapSensor AfterConnected()");

            //连接断开，复位 脉冲计数。。。,时间计数
            isReadyContext.Reset();
            sysTickContext.Reset();

            //core.GetIStatus((asyncContext, retData) =>
            //{
            //    var istatus = (UInt16)retData;
            //    IStatus = istatus;
            //    isReadyContext.isReadyGetIn = true;
            //}, null);
        }
        #region ICapSensor2Client
        /// <summary>
        /// 连接
        /// </summary>
        public void Connect()
        {
            Connect(Addr);
        }
        public void Disconnect()
        {
            IsRunning = false;
            if (comm != null)
            {
                comm.Stop();
                disposeComm();
                comm = null;
            }
        }

        void createComm()
        {
            if (Addr.ToLower().StartsWith("com"))
            {
                var _comm = new GComm_SerialPort();
                _comm.Init(new System.IO.Ports.SerialPort()
                {
                    BaudRate = 115200,
                    DataBits = 8,
                    Parity = System.IO.Ports.Parity.None,
                    StopBits = System.IO.Ports.StopBits.One
                });
                //_comm.logger = logger;
                comm = _comm;
            }
            else
            {
                var _comm = new GComm_TcpClient();
                //_comm.logger = logger;
                comm = _comm;
            }
            comm.DataReceived += Comm_DataReceived;
            comm.PropertyChanged += Comm_PropertyChanged;
            core.TimeOutEvent += Core_TimeOutEvent;
            core.ResetMsg();
        }

        private void Core_TimeOutEvent(object sender)
        {
            //回复超时，可能连接断开， 重新连接
            ReConnect();
        }

        void disposeComm()
        {
            comm.DataReceived -= Comm_DataReceived;
            comm.PropertyChanged -= Comm_PropertyChanged;
            core.TimeOutEvent -= Core_TimeOutEvent;
            core.ResetMsg();
        }
        private void Comm_DataReceived(IGeneralComm sender, byte[] msg)
        {
            core.RecMsg(msg);
        }
        private void Comm_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName == nameof(comm.IsConnected))
            {
                if (!comm.IsConnected)
                {
                    logger.Error($"comm 连接断开 {comm.ErrMsg}");
                }
                //连接状态改变， 复位通信状态
                core.ResetMsg();
            }
        }

        /// <summary>
        /// 连接 addr
        /// </summary>
        /// <param name="addr"></param>
        public void Connect(string addr)
        {
            IsRunning = true;

            Addr = addr;
            if (comm == null)
            {
                createComm();
            }
            else
            {
                comm.Stop();
                //只要comm.Stop()  IsConnected 就会为 False

                if ((Addr.ToLower().StartsWith("com") && (comm is GComm_SerialPort)) ||
                    (!Addr.ToLower().StartsWith("com") && (comm is GComm_TcpClient)))
                {
                    //不用重建
                }
                else
                {
                    disposeComm();
                    createComm();
                }
            }

            comm.Addr = addr;
            comm.Start();
        }

        /// <summary>
        /// 重连
        /// </summary>
        public void ReConnect()
        {
            Connect();
        }

        #endregion




        #region Load Save




        string filePath = "capsensor.json";

        public bool Save()
        {
            return CapSensor2JsonDb.Save(filePath, this);
        }
        public bool Load()
        {
            bool ret = CapSensor2JsonDb.Load(filePath, this);
            return ret;
        }
        public bool Save(string filePath)
        {
            if (filePath != null)
                this.filePath = filePath;
            return Save();
        }
        public bool Load(string filePath)
        {
            if (filePath != null)
                this.filePath = filePath;

            return Load();

        }
        #endregion
    }

    public class CapSensor2JsonDb
    {
        public static bool Save(string filePath, CapSensor2 src)
        {

            if (string.IsNullOrEmpty(filePath))
                return false;
            var p = new CapSensor2JsonDb()
            {
                Addr = src.Addr
            };
            try
            {
                File.WriteAllText(filePath, JsonConvert.SerializeObject(p, Formatting.Indented));
                return true;
            }
            catch
            {
                //异常，没有json 编码失败
                return false;
            }
        }
        public static bool Load(string filePath, CapSensor2 src)
        {
            if (string.IsNullOrEmpty(filePath))
                return false;

            try
            {
                if (File.Exists(filePath))
                {
                    string json = File.ReadAllText(filePath);

                    var p = JsonConvert.DeserializeObject<CapSensor2JsonDb>(json);
                    src.Addr = p.Addr;
                    return true;
                }
            }
            catch
            {
                //异常，没有json 解码失败
            }
            return false;
        }

        public string Addr = "192.168.251.12:20020";
    }


    /// <summary>
    ///  systick 转 DateTime
    ///  不断把 从flyad7 获取的 systick 放入。  以每次systick 的差 加上 上一次 dt_1st, 得到时间。
    ///  如： 当前 获得 systick1, 之前是 systick0,   所以 dt = dt_1st+ (systick1-systick0);  dt_1st=dt;
    ///  
    /// 第1个 dt_1st, 是 从 DateTime.Now. 获取的。
    /// 
    /// 不会管 当前DateTime.Now 与 dt 差多少。 所以如果 flyad7 时间比较慢，可能会出现 3天后， DateTime.Now 与 dt 差了 1个小时。
    /// </summary>
    class SysTickContext
    {
        byte systick_1st = 0;
        DateTime dt_1st = DateTime.MinValue;

        /// <summary>
        /// 
        /// </summary>
        public event Action BeReseted;
        public void Reset()
        {
            dt_1st = DateTime.MinValue;
        }
        /// <summary>
        /// 输入从 flyad7 获取的 timemark, 转换为 本地 DateTime
        /// </summary>
        /// <param name="systick">只有低7位有效</param>
        /// <returns></returns>
        public DateTime ToDateTime(byte systick)
        {
            if (dt_1st == DateTime.MinValue)
            {
                dt_1st = DateTime.Now;
                //毫秒级别取整
                long tick = dt_1st.Ticks % TimeSpan.TicksPerMillisecond;
                dt_1st -= TimeSpan.FromTicks(tick);

                systick_1st = systick;
                return dt_1st;
            }

            int ms = SystickSub(systick, systick_1st);
            if (ms > 60)//记录与新的timemark 相差60ms，掉包很严重
            {
                //被复位了
                BeReseted?.Invoke();
            }

            DateTime dt = dt_1st.AddMilliseconds(ms);

            systick_1st = systick;
            dt_1st = dt;
            return dt;
        }

        /// <summary>
        /// 8位 systick1-systick2
        /// </summary>
        /// <param name="systick1"></param>
        /// <param name="systick2"></param>
        /// <returns></returns>
        int SystickSub(byte systick1, byte systick2)
        {
            if (systick1 >= systick2)
                return systick1 - systick2;
            else
            {
                return systick1 + 0x100 - systick2;
            }
        }
    }

    class IsReadyContext : INotifyPropertyChanged
    {
        public event PropertyChangedEventHandler PropertyChanged;

        public bool isReadyGetIn { get; set; }

        public bool IsReady { get; private set; }

        void updateIsReady()
        {

            if (isReadyGetIn == false)
            {
                IsReady = false;
                return;
            }

            IsReady = true;
            return;
        }

        public IsReadyContext()
        {
            this.PropertyChanged += IsReadyContext_PropertyChanged;
        }

        private void IsReadyContext_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName != nameof(IsReady))
            {
                updateIsReady();
            }
        }
        public void Reset()
        {

            isReadyGetIn = false;
        }
    }
}
