﻿using GeneralGommunication;
using System;
using System.Collections.Generic;
using System.Linq;

namespace CapSensorBase
{
    public class CapSensorCore : Dev7E, ICapSensorCore
    {

        public event CoreEvent_HzMeasuredOutEventHandler HzMeasuredOut;
        public event CoreEvent_IStatusChangedEventHandler IStatusChanged;
        public event CoreEvent_DataPushedEventHandler DataPushed;
        public event CoreEvent_TempMeasuredOutEventHandler TempMeasuredOut;


        public CapSensorCore()
        {
            logger = NLog.LogManager.GetCurrentClassLogger();
            PrefixIndex = 1;
        }

        protected override void ParsePackAfterCheckCRC8(byte[] pack)
        {
            byte B0 = pack[0];

            // PT = 0--数据包(推送包)，1--功能包
            bool PT = Misc.MyBase.CHECKBIT(B0, 7);

            if (!PT)
            {
                //数据包
                ParseDataPack(pack);
            }
            else
            {
                ParseFuncPack(pack);
            }
        }

        protected override List<byte> GetSendPack(List<byte> datas)
        {
            datas.Insert(0, 0x80);//插入B0
            return base.GetSendPack(datas);
        }

        void ParseDataPack(byte[] buf)
        {
            if (buf.Count() < 3)
            {
                ErrCnt++;
                return;//异常
            }
            byte B0 = buf[0];

            //数据包的顺序码。 7位，肯定是1，2，3，4 排序。 如果少了一个。肯定少了一个数据包
            byte packno = B0;
            Misc.MyBase.CLEARBIT(ref packno, 7);

            //数据包类型
            byte B1 = buf[1];

            DataPushed?.Invoke(this, new CoreEvent_DataPushedEventArgs()
            {
                packno = packno,
                dataType = (char)B1
            });


            //8位 systick 单位1ms
            byte systick = buf[2];

            int index = 3;

            switch (B1)
            {
                case (byte)'F':
                    {
                        //CH(4b)+serial(4b)+data(2B)
                        byte channel = (byte)(buf[index] >> 4);
                        byte serial = (byte)(buf[index] & 0xF);
                        index++;
                        UInt16 data = BitConverter.ToUInt16(buf, index);
                        index += 2;

                        HzMeasuredOut?.Invoke(this, new CoreEvent_HzMeasuredOutEventArgs()
                        {
                            systick = systick,
                            channel = channel,
                            serial = serial,
                            data = data
                        });
                    }
                    break;
                case (byte)'T':
                    {
                        //temperture(2B)+pwm(2B)

                        UInt16 temp = BitConverter.ToUInt16(buf, index);
                        index += 2;
                        UInt16 pwm = BitConverter.ToUInt16(buf, index);
                        index += 2;

                        double temperature = temp / 100.0;
                        var args = new CoreEvent_TempMeasuredOutEventArgs()
                        {
                            systick = systick,
                            temperature = temperature,
                            pwm = pwm
                        };

                        TempMeasuredOut?.Invoke(this, args);
                    }
                    break;
                case (byte)'I':
                    {
                        //0(1B)+istatus(2B)+changed(2B)
                        index++;
                        UInt16 istatus = BitConverter.ToUInt16(buf, index);
                        index += 2;
                        UInt16 ichanged = BitConverter.ToUInt16(buf, index);
                        index += 2;

                        IStatusChanged?.Invoke(this, new CoreEvent_IStatusChangedEventArgs()
                        {
                            systick = systick,
                            istatus = istatus,
                            ichanged = ichanged
                        });
                    }
                    break;
            }
        }



        /// <summary>
        /// frequency divider 设置分频
        /// </summary>
        public void SetFreqDiv(UInt16 fd, CallBackHandler asyncDelegate, object asyncContext)
        {
            List<byte> datas = new List<byte>();
            datas.AddRange(BitConverter.GetBytes(fd));

            var datasObj = new { fd };

            COMMREQ commReq = new COMMREQ()
            {
                Prefix = COMMREQ.ToPrefix("MPA")
            };

            //放入 交易队列
            AddTran(
                new COMMREQ_Transaction()
                {
                    commReq = commReq,
                    asyncDelegate = asyncDelegate,
                    asyncContext = asyncContext,
                    datas = datas.ToArray(),
                    datasObj = datasObj
                });
        }

        /// <summary>
        /// 设置 是否自动计算分频数
        /// </summary>
        /// <param name="isAuto"></param>
        public void SetAutoFreqDiv(bool isAuto, CallBackHandler asyncDelegate, object asyncContext)
        {
            List<byte> datas = new List<byte>();
            datas.Add((byte)(isAuto ? '1' : '0'));

            var datasObj = new { isAuto };

            COMMREQ commReq = new COMMREQ()
            {
                Prefix = COMMREQ.ToPrefix("MPB")
            };

            //放入 交易队列
            AddTran(
                new COMMREQ_Transaction()
                {
                    commReq = commReq,
                    asyncDelegate = asyncDelegate,
                    asyncContext = asyncContext,
                    datas = datas.ToArray(),
                    datasObj = datasObj
                });
        }

        /// <summary>
        /// 强制读取分频数
        /// </summary>
        [CallBack(typeof(GetFreqDivResponse))]
        public void GetFreqDiv(CallBackHandler asyncDelegate, object asyncContext)
        {
            COMMREQ commReq = new COMMREQ()
            {
                Prefix = COMMREQ.ToPrefix("Mp"),
                ResponseLen = 3,
                ParseFuncPack = (pack, dataIdx) =>
                {
                    var freqDiv = BitConverter.ToUInt16(pack, dataIdx);
                    dataIdx += 2;
                    var isFdAuto = pack[dataIdx] == '1';
                    dataIdx++;

                    return new GetFreqDivResponse()
                    {
                        freqDiv = freqDiv,
                        isFdAuto = isFdAuto
                    };
                }
            };

            //放入 交易队列
            AddTran(
                new COMMREQ_Transaction()
                {
                    commReq = commReq,
                    asyncDelegate = asyncDelegate,
                    asyncContext = asyncContext
                });
        }
        #region PID
        public void SetP(float kp, CallBackHandler asyncDelegate, object asyncContext)
        {
            List<byte> datas = new List<byte>();
            datas.AddRange(BitConverter.GetBytes(kp));

            var datasObj = new { kp };

            COMMREQ commReq = new COMMREQ()
            {
                Prefix = COMMREQ.ToPrefix("PPP")
            };

            //放入 交易队列
            AddTran(
                new COMMREQ_Transaction()
                {
                    commReq = commReq,
                    asyncDelegate = asyncDelegate,
                    asyncContext = asyncContext,
                    datas = datas.ToArray(),
                    datasObj = datasObj
                });
        }
        public void SetI(float ki, CallBackHandler asyncDelegate, object asyncContext)
        {
            List<byte> datas = new List<byte>();
            datas.AddRange(BitConverter.GetBytes(ki));

            var datasObj = new { ki };

            COMMREQ commReq = new COMMREQ()
            {
                Prefix = COMMREQ.ToPrefix("PPI")
            };

            //放入 交易队列
            AddTran(
                new COMMREQ_Transaction()
                {
                    commReq = commReq,
                    asyncDelegate = asyncDelegate,
                    asyncContext = asyncContext,
                    datas = datas.ToArray(),
                    datasObj = datasObj
                });
        }

        public void SetD(float kd, CallBackHandler asyncDelegate, object asyncContext)
        {
            List<byte> datas = new List<byte>();
            datas.AddRange(BitConverter.GetBytes(kd));

            var datasObj = new { kd };

            COMMREQ commReq = new COMMREQ()
            {
                Prefix = COMMREQ.ToPrefix("PPD")
            };

            //放入 交易队列
            AddTran(
                new COMMREQ_Transaction()
                {
                    commReq = commReq,
                    asyncDelegate = asyncDelegate,
                    asyncContext = asyncContext,
                    datas = datas.ToArray(),
                    datasObj = datasObj
                });
        }

        public void SetT(double target, CallBackHandler asyncDelegate, object asyncContext)
        {
            List<byte> datas = new List<byte>();

            UInt16 t = (UInt16)(target * 100);
            datas.AddRange(BitConverter.GetBytes(t));

            var datasObj = new { target };

            COMMREQ commReq = new COMMREQ()
            {
                Prefix = COMMREQ.ToPrefix("PPT")
            };

            //放入 交易队列
            AddTran(
                new COMMREQ_Transaction()
                {
                    commReq = commReq,
                    asyncDelegate = asyncDelegate,
                    asyncContext = asyncContext,
                    datas = datas.ToArray(),
                    datasObj = datasObj
                });
        }

        [CallBack(typeof(float))]
        public void GetP(CallBackHandler asyncDelegate, object asyncContext)
        {
            COMMREQ commReq = new COMMREQ()
            {
                Prefix = COMMREQ.ToPrefix("PpP"),
                ResponseLen = 4,
                ParseFuncPack = (pack, dataIdx) =>
                {
                    return BitConverter.ToSingle(pack, dataIdx);
                }
            };

            //放入 交易队列
            AddTran(
                new COMMREQ_Transaction()
                {
                    commReq = commReq,
                    asyncDelegate = asyncDelegate,
                    asyncContext = asyncContext
                });
        }

        [CallBack(typeof(float))]
        public void GetI(CallBackHandler asyncDelegate, object asyncContext)
        {
            COMMREQ commReq = new COMMREQ()
            {
                Prefix = COMMREQ.ToPrefix("PpI"),
                ResponseLen = 4,
                ParseFuncPack = (pack, dataIdx) =>
                {
                    return BitConverter.ToSingle(pack, dataIdx);
                }
            };

            //放入 交易队列
            AddTran(
                new COMMREQ_Transaction()
                {
                    commReq = commReq,
                    asyncDelegate = asyncDelegate,
                    asyncContext = asyncContext
                });
        }

        [CallBack(typeof(float))]
        public void GetD(CallBackHandler asyncDelegate, object asyncContext)
        {
            COMMREQ commReq = new COMMREQ()
            {
                Prefix = COMMREQ.ToPrefix("PpD"),
                ResponseLen = 4,
                ParseFuncPack = (pack, dataIdx) =>
                {
                    return BitConverter.ToSingle(pack, dataIdx);
                }
            };

            //放入 交易队列
            AddTran(
                new COMMREQ_Transaction()
                {
                    commReq = commReq,
                    asyncDelegate = asyncDelegate,
                    asyncContext = asyncContext
                });
        }

        [CallBack(typeof(double))]
        public void GetT(CallBackHandler asyncDelegate, object asyncContext)
        {
            COMMREQ commReq = new COMMREQ()
            {
                Prefix = COMMREQ.ToPrefix("PpT"),
                ResponseLen = 2,
                ParseFuncPack = (pack, dataIdx) =>
                {
                    double t = BitConverter.ToUInt16(pack, dataIdx);

                    return t / 100.0;
                }
            };

            //放入 交易队列
            AddTran(
                new COMMREQ_Transaction()
                {
                    commReq = commReq,
                    asyncDelegate = asyncDelegate,
                    asyncContext = asyncContext
                });
        }
        #endregion

        /// <summary>
        /// 设置PWM 自动
        /// </summary>
        /// <param name="on"></param>
        public void SetPwmAuto(bool on, CallBackHandler asyncDelegate, object asyncContext)
        {
            //TODO, 究竟返回什么？
            COMMREQ commReq = new COMMREQ()
            {
                Prefix = COMMREQ.ToPrefix(on ? "PA1" : "PA0")
            };

            //放入 交易队列
            AddTran(
                new COMMREQ_Transaction()
                {
                    commReq = commReq,
                    asyncDelegate = asyncDelegate,
                    asyncContext = asyncContext
                });
        }
        /// <summary>
        /// 设置PWM 值
        /// </summary>
        /// <param name="pwm"></param>
        public void SetPwm(int pwm, CallBackHandler asyncDelegate, object asyncContext)
        {
            List<byte> datas = new List<byte>();

            UInt16 t = (UInt16)(pwm);

            datas.AddRange(BitConverter.GetBytes(t));

            var datasObj = new { pwm };

            COMMREQ commReq = new COMMREQ()
            {
                Prefix = COMMREQ.ToPrefix("PAV")
            };

            //放入 交易队列
            AddTran(
                new COMMREQ_Transaction()
                {
                    commReq = commReq,
                    asyncDelegate = asyncDelegate,
                    asyncContext = asyncContext,
                    datas = datas.ToArray(),
                    datasObj = datasObj
                });
        }

        [CallBack(typeof(UInt16))]
        public void GetIStatus(CallBackHandler asyncDelegate, object asyncContext)
        {
            COMMREQ commReq = new COMMREQ()
            {
                Prefix = COMMREQ.ToPrefix("IGI"),
                ResponseLen = 2,
                ParseFuncPack = (pack, dataIdx) => { return BitConverter.ToUInt16(pack, dataIdx); }
            };

            //放入 交易队列
            AddTran(
                new COMMREQ_Transaction()
                {
                    commReq = commReq,
                    asyncDelegate = asyncDelegate,
                    asyncContext = asyncContext
                });
        }

    }
}
