﻿using CapSensorBase;
using FlyADBase;
using FObjBase;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace FLY.Thick.FilmCasting.Cap.Server
{
    /// <summary>
    /// 设备时间同步
    /// </summary>
    public class DevSync : INotifyPropertyChanged
    {
        public event PropertyChangedEventHandler PropertyChanged;

        /// <summary>
        /// 同步间隔分钟数
        /// </summary>
        const int SyncIntervalMin = 10;
        /// <summary>
        /// Ad盒 设备序号
        /// </summary>
        const int FlyAd_DevIdx = 0;

        /// <summary>
        /// 电容传感器 设备序号
        /// </summary>
        const int CapSensor_DevIdx = 1;

        /// <summary>
        /// 电容传感器同步线输入
        /// </summary>
        const int CapSensor_InNo_Sync = 0;
        /// <summary>
        /// Ad盒X0作为同步线输入
        /// </summary>
        const int FlyAd_InNo_Sync = 0;
        /// <summary>
        /// Ad盒Y0作为同步线输输出
        /// </summary>
        const int FlyAd_OutNo_Sync = 0;

        /// <summary>
        /// 上一次同步时间
        /// </summary>
        public DateTime SyncTime { get; private set; }
        /// <summary>
        /// 同步成功
        /// </summary>
        public bool IsSyncSuccess { get; private set; }

        /// <summary>
        /// 从系统时间偏移
        /// </summary>
        public TimeSpan SlaveSysTimeOffset { get; private set; }

        /// <summary>
        /// 从系统时间偏移 单位ms
        /// </summary>
        public int SlaveSysTimeOffsetMs => (int)SlaveSysTimeOffset.TotalMilliseconds;
        /// <summary>
        /// 下次同步剩余秒数
        /// </summary>
        public int NextSyncSurplusSec { get; private set; }


        FlyAD7 flyAd;
        CapSensor3 capSensor;

        /// <summary>
        /// AD盒 同步时的 系统时间
        /// </summary>
        DateTime[] syncTimes;

        enum SyncStatus
        {
            /// <summary>
            /// 等待2个AD盒都连接成功,然后 主设备 同步输出位 =0
            /// </summary>
            WaitForConnected,

            /// <summary>
            /// 主设备 同步输出位 =0
            /// </summary>
            SyncSignLow,

            /// <summary>
            /// 等待成功
            /// </summary>
            WaitForSuccess,

            /// <summary>
            /// 等待下次同步
            /// </summary>
            WaitForNextSync
        }
        SyncStatus Status;
        Stopwatch stopwatch = new Stopwatch();
        
        public DevSync()
        {

        }

        public void Init(FlyAD7 flyAd, CapSensor3 capSensor)
        {
            this.flyAd = flyAd;
            this.capSensor = capSensor;

            syncTimes = new DateTime[2];
            SlaveSysTimeOffset = TimeSpan.Zero;

            flyAd.IStatusChangedEvent += FlyAd_IStatusChangedEvent; ;
            flyAd.PropertyChanged += FlyAd_PropertyChanged;
            capSensor.PropertyChanged += CapSensor_PropertyChanged;
            capSensor.IStatusChanged += CapSensor_IStatusChanged;
            capSensor.SystickReseted += CapSensor_SystickReseted;
            Status = SyncStatus.WaitForConnected;
            PollModule.Current.Poll_Config(OnPoll, TimeSpan.FromSeconds(1));
        }

        private void CapSensor_SystickReseted(object sender)
        {
            Sync();
        }




        /// <summary>
        /// 1秒一次
        /// </summary>
        private void OnPoll()
        {
            switch (Status)
            {
                case SyncStatus.WaitForConnected:
                    if (!flyAd.IsReady)
                        break;
                    if (!capSensor.IsReady)
                        break;

                    for (int i = 0; i < syncTimes.Count(); i++)
                        syncTimes[i] = DateTime.MinValue;
                    

                    //拉低
                    //Y0 = 0, AD盒X0, 电容传感器X0会同时收到信号1->0信号
                    flyAd.SetOutputBit(FlyAd_OutNo_Sync, false);
                    Status = SyncStatus.SyncSignLow;
                    break;
                case SyncStatus.SyncSignLow:
                    {
                        //拉高
                        //Y0 = 1, 恢复Y0状态，准备下次
                        flyAd.SetOutputBit(FlyAd_OutNo_Sync, true);

                        stopwatch.Restart();
                        Status = SyncStatus.WaitForSuccess;
                    }
                    break;
                case SyncStatus.WaitForSuccess:
                    {
                        if (IsSyncSuccess)
                        {
                            stopwatch.Restart();
                            Status = SyncStatus.WaitForNextSync;
                            SyncTime = DateTime.Now;
                            break;
                        }
                        else
                        {
                            if (stopwatch.Elapsed < TimeSpan.FromSeconds(3))
                                break;//等3秒
                            //异常5秒都没反应，可能没接线，重试
                            Status = SyncStatus.WaitForConnected;
                        }
                    }
                    break;
                case SyncStatus.WaitForNextSync:
                    {
                        NextSyncSurplusSec = (int) (TimeSpan.FromMinutes(SyncIntervalMin) - stopwatch.Elapsed).TotalSeconds;
                        if (NextSyncSurplusSec > 0)
                            break;
                        Sync();
                    }
                    break;
            }
        }

        /// <summary>
        /// 立刻同步
        /// </summary>
        public void Sync()
        {
            if (Status == SyncStatus.WaitForNextSync)
            {
                NextSyncSurplusSec = 0;
                stopwatch.Stop();
                Status = SyncStatus.WaitForConnected;
            }
        }


        private void FlyAd_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName == nameof(flyAd.IsReady))
            {
                if (!flyAd.IsReady)
                {
                    IsSyncSuccess = false;
                    for (int i = 0; i < syncTimes.Count(); i++)
                        syncTimes[i] = DateTime.MinValue;

                    Status = SyncStatus.WaitForConnected;
                }
            }
        }
        bool isInWaitForSuccess() 
        {
            if (Status == SyncStatus.WaitForSuccess || Status == SyncStatus.SyncSignLow) 
            {
                return true;
            }
            return false;
        }
        private void CapSensor_IStatusChanged(object sender, CapSensorBase.IStatusChangedEventArgs e)
        {
            if (!isInWaitForSuccess())
                return;

            if (!Misc.MyBase.CHECKBIT(e.ichanged, CapSensor_InNo_Sync))
                return;


            if (!Misc.MyBase.CHECKBIT(e.istatus, CapSensor_InNo_Sync))
                return;
            
            if (syncTimes[CapSensor_DevIdx] != DateTime.MinValue)
                return;

            //收到信号了
            syncTimes[CapSensor_DevIdx] = e.time;

            updateSlaveSysTimeOffset();
        }

        private void CapSensor_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName == nameof(CapSensor3.IsReady))
            {
                if (!capSensor.IsReady)
                {
                    IsSyncSuccess = false;
                    for (int i = 0; i < syncTimes.Count(); i++)
                        syncTimes[i] = DateTime.MinValue;

                    Status = SyncStatus.WaitForConnected;
                }
            }
        }
        
        private void FlyAd_IStatusChangedEvent(object sender, FlyADBase.IStatusChangedEventArgs e)
        {
            if (!isInWaitForSuccess())
                return;

            if (!Misc.MyBase.CHECKBIT(e.IChanged, FlyAd_InNo_Sync))
                return;


            if (!Misc.MyBase.CHECKBIT(e.IStatus, FlyAd_InNo_Sync))
                return;

            if (syncTimes[FlyAd_DevIdx] != DateTime.MinValue)
                return;

            //收到信号了
            syncTimes[FlyAd_DevIdx] = e.Time;

            updateSlaveSysTimeOffset();
        }


        bool updateSlaveSysTimeOffset()
        {
            if (syncTimes.Any(t => t == DateTime.MinValue))
                return false;

            SlaveSysTimeOffset = syncTimes[0] - syncTimes[1];
            IsSyncSuccess = true;
            return true;
        }
    }
}
