﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;

namespace FLY.Simulation.Blowing
{
    public class GageAD : ISimulationGageAD, INotifyPropertyChanged
    {
        /// <summary>
        /// 1脉冲 = ? mm
        /// </summary>
        public double Mmpp { get; set; } = 0.075;

        /// <summary>
        /// 机架总长 mm
        /// </summary>
        public int TotalLengthMm { get; set; } = 845;

        /// <summary>
        /// 原点信号位置
        /// </summary>
        public int OrgSignMm { get; set; } = 30;

        /// <summary>
        /// 膜宽,mm
        /// </summary>
        public int FilmWidth { get; set; }

        /// <summary>
        /// 膜开始位置, mm
        /// </summary>
        public int FilmBegin { get; set; }

        /// <summary>
        /// 传感器直径，单位mm
        /// </summary>
        public int SenserWidth { get; set; }

        #region 探头气动装置
        /// <summary>
        /// 下压按钮被触发， 电箱的按钮<para/>
        /// 接X4, 没触发时X4=1, 触发时X4=0
        /// </summary>
        public bool IsPushDownButtonTriggered { get; set; }

        /// <summary>
        /// 安装在探头前方的保护装置被触发 <para/>
        /// 接X5, 正常时X5=1, 当膜非常厚，使保护装置抬起 触发， X5=0
        /// </summary>
        public bool IsSensorProtectionTrggered { get; set; }
        #endregion

        /// <summary>
        /// 下压状态
        /// </summary>
        bool isInPushDown = true;
        bool isSyncIn;

        /// <summary>
        /// AD盒 编码器位置
        /// </summary>
        int position;

        Blowing mBlowing;
        CurveCollection curve;

        int[] sensorWeights;
        int sensorWeightSum;

        const int HzOffsetMax = 100000;
        const double HzOffsetScaleA = 4 / 120;
        const double HzOffsetScaleMax = 30.0 / 120;
        const double HzOffsetScaleMin = 8 / 120;

        DateTime now = DateTime.MinValue;
        DateTime hzOffsetDt = DateTime.MinValue;
        double currHzOffset = 0;
        bool lastIsInPushDown = false;
        double currHzOffsetScale = HzOffsetScaleMin;

        public GageAD()
        {
            curve = new CurveCollection();
        }
        public void Init(Blowing blowing)
        {
            mBlowing = blowing;
            FilmWidth = (int)mBlowing.FilmWidth;
            FilmBegin = ( TotalLengthMm  - FilmWidth) / 2;

            SenserWidth = 15;

            updateSensorWeights();
        }
        void updateSensorWeights() 
        {
            sensorWeights = new int[SenserWidth];

            for (int i = 0; i < sensorWeights.Count(); i++) 
            {
                int mid_index = sensorWeights.Count() / 2;
                double weight = Math.Sqrt(Math.Pow(mid_index, 2) - Math.Pow(mid_index - i, 2));
                sensorWeights[i] = (int)weight;
            }

            sensorWeightSum = sensorWeights.Sum();
        }
        public void OnPoll(DateTime now)
        {
            this.now = now;
            //NewAirDatas();
            mBlowing.OnPoll(now);
            OnPoll_HzOffset();

        }
        /// <summary>
        /// 探头抬升，会导致Hz变大<para/>
        /// 探头下压，会导致Hz变小<para/>
        /// 刚开始Hz变化大，后面变化慢，整个过程需要120s才变成完成<para/>
        /// </summary>
        void OnPoll_HzOffset()
        {
            if (hzOffsetDt == DateTime.MinValue)
            {
                hzOffsetDt = now;
                lastIsInPushDown = isInPushDown;
                return;
            }
            if (lastIsInPushDown != isInPushDown)
            {
                lastIsInPushDown = isInPushDown;

                //气缸状态一变化， 变化率就会变为最大
                currHzOffsetScale = HzOffsetScaleMax;
            }



            double sec = (now - hzOffsetDt).TotalSeconds;
            hzOffsetDt = now;

            //变化率逐秒递减，直达最小变化率
            if (currHzOffsetScale > HzOffsetScaleMin)
                currHzOffsetScale -= sec * HzOffsetScaleA;
            if (currHzOffsetScale < HzOffsetScaleMin)
                currHzOffsetScale = HzOffsetScaleMin;

            //Hz变化量改变
            if (isInPushDown)
            {
                //Hz变化量不断减速
                if (currHzOffset > 0)
                {
                    currHzOffset -= currHzOffset * sec * currHzOffsetScale;
                    if (currHzOffset < 2)
                    {
                        currHzOffset = 0;
                    }
                }
            }
            else
            {
                //Hz变化量不断加大
                if (currHzOffset < HzOffsetMax)
                {
                    currHzOffset += (HzOffsetMax - currHzOffset) * sec * currHzOffsetScale;
                    if (currHzOffset > HzOffsetMax - 2)
                    {
                        currHzOffset = HzOffsetMax;
                    }
                }
            }
        }

        /// <summary>
        /// 获取AD值
        /// </summary>
        /// <returns></returns>
        public int GetAD()
        {
            return 0;
        }

        public int GetHz()
        {
            int ad = GetAD((int)(position * Mmpp));
            return (int)(ad + currHzOffset);
        }
        public int GetAD(int mm)
        {
            return GetAD_1(mm);
            //return GetAD_2(mm);
        }

        Random r = new Random();

        public event PropertyChangedEventHandler PropertyChanged;

        int GetAD_1(int mm)
        {
            int data = r.Next(30 * 2) - 30;// 单位 0.01um， ±0.3um波动

            if ((mm >= FilmBegin) && (mm < (FilmBegin + FilmWidth)))
            {
                int position_mm = mm - FilmBegin;
                FilmData fd = mBlowing.GetData(position_mm);
                if (fd != null)
                {
                    data += fd.data1 + fd.data2;
                }
            }

            //求平均值
            double thick = data / 100.0;
            return curve.Value2Ad(thick, AD2ValueFlag.NoRevised);
        }

        /// <summary>
        /// 带卷积的数据混合
        /// </summary>
        /// <param name="mm"></param>
        /// <returns></returns>
        int GetAD_2(int mm)
        {
            if (SenserWidth != sensorWeights.Count()) {
                updateSensorWeights();
            }

            List<int> datas = new List<int>();
            int weightsum = 0;
            datas.Add(GetAD_1(mm) * sensorWeights[SenserWidth / 2]);
            weightsum += sensorWeights[SenserWidth / 2];


            for (int i = 0; i < SenserWidth/2; i+=2)
            {
                datas.Add(GetAD_1(mm - i) * sensorWeights[SenserWidth / 2 - i ]);
                datas.Add(GetAD_1(mm + i) * sensorWeights[SenserWidth / 2 + i ]);

                weightsum += sensorWeights[SenserWidth / 2 - i];
                weightsum += sensorWeights[SenserWidth / 2 + i];
            }
            //求平均值
            return (int)(datas.Sum() / weightsum);
        }
        /// <summary>
        /// 获取输入口
        /// </summary>
        /// <returns></returns>
        public UInt16 GetInput()
        {
            UInt16 istatus = 0x0fff;

            if (!mBlowing.IsShieldI9)
            {
                if (mBlowing.IsTurnSign0)
                    Misc.MyBase.CLEARBIT(ref istatus, IODefinition.IN_TurnSign0);
            }

            if (mBlowing.IsTurnSign1)
                Misc.MyBase.CLEARBIT(ref istatus, IODefinition.IN_TurnSign1);


            if (mBlowing.IsRollSign)
                Misc.MyBase.CLEARBIT(ref istatus, IODefinition.IN_Roll);

            if (isSyncIn)
                Misc.MyBase.CLEARBIT(ref istatus, IODefinition.IN_Sync);

            if (IsPushDownButtonTriggered)
                Misc.MyBase.CLEARBIT(ref istatus, IODefinition.IN_PushDownButton);

            if (!IsSensorProtectionTrggered)
                Misc.MyBase.CLEARBIT(ref istatus, IODefinition.IN_SensorProtection);

            return istatus;
        }


        public int GetPosition2()
        {
            double p = mBlowing.GlobalAngle / 360;
            return (int)(mBlowing.PosOfR * p);
        }
        public int GetSpeed2()
        {
            return (int)(mBlowing.PosOfR * mBlowing.CurrAngleVelocity / 360);
        }

        public void SetOutput(ushort output)
        {
            isSyncIn = !Misc.MyBase.CHECKBIT(output, IODefinition.OUT_Sync);

            isInPushDown = !Misc.MyBase.CHECKBIT(output, IODefinition.OUT_PushDown);

        }
        public bool GetSyncIn() 
        {
            return isSyncIn;
        }


        /// <summary>
        /// AD盒 移动了，设置编码器1位置
        /// </summary>
        /// <param name="position"></param>
        public void SetPosition(int position) 
        {
            this.position = position;
        }
    }
}


